/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;


/**
 * Declared optional parameters of a function. It's a part of function's signature representing
 * type and position of a list of all optional parameters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class OptParameters {

    /**
     * Flag informing whether the object has been configured using {@link #configure(int)}
     * method.
     */
    private boolean isConfigured;

    /**
     * Nonnegative index of first optional parameter on a list of parameters declared by
     * function's signature the object is bound to. Value of this field is provided by the
     * signature object.
     */
    private int startIndex;

    /**
     * Constructs object representing function's declared optional parameters.
     * <p>
     * Note: Before the object representing parameters can be used, it must be bound to a
     * function's signature by passing it to {@link FunctionSignature} constructor.
     */
    protected OptParameters() { }

    /**
     * Configures this object representing function's declared optional parameters by providing
     * its index on a list of declared parameters. Every <code>OptionalParameters</code> must be
     * configured before it can be used. Signature object that <code>OptParameters</code>
     * instance is bound to is responsible for executing configuration exactly once.
     *
     * @param paramsStartIndex
     *     nonnegative position of the parameters in a list of declared parameters contained in
     *     the signature
     */
    final void configure(int paramsStartIndex) {
        if (isConfigured) {
            throw new IllegalStateException("The object may be configured only once");
        }
        if (paramsStartIndex < 0) {
            throw new IllegalArgumentException("The object's index must not be negative");
        }

        isConfigured = true;
        startIndex = paramsStartIndex;
    }

    /**
     * Returns index of first optional parameter on a list of parameters declared by
     * function's signature the object is bound to.
     * <p>
     * Note: Before the method can be used, this object must be bound to a function's signature
     * by passing it to {@link FunctionSignature} constructor.
     *
     * @return index of first optional parameter on a list of declared parameters
     */
    public final int getStartIndex() {
        if (!isConfigured) {
            throw new IllegalStateException(
                "Parameters object must be configured before it can be used");
        }
        return startIndex;
    }

    /**
     * Returns type of optional parameters represented by the object
     *
     * @return type of the parameters represented by the object
     */
    public abstract Type getType();

}
