/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core;

import java.text.SimpleDateFormat;
import java.util.Date;
import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.utils.LongUtils;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;
import pl.chyla.wyjec.utils.SimpleCalendar;


/**
 * Container holding single value of one of possible types. At any time an instance of this
 * class allows accessing value previously set and querying its type.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Value {

    /**
     * Date format to use when generating string representation of date value.
     */
    private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss";

    /**
     * Boolean value contained in the object (for use only if its current type is BOOLEAN)
     */
    private boolean booleanValue;

    /**
     * Integer value contained in the object (for use only if its current type is INTEGER)
     */
    private long integerValue;

    /**
     * Array used to store characters of value of type STRING. It's used by private RawString
     * object representing STRING value ({@link #stringValue}) as storage for string's
     * characters. It's possible to change storage of <code>stringValue</code> to some other
     * array using {@link #setString(ReadableRawString)} method but
     * <code>origStringChars</code> never gets lost and will be used again as storage after
     * resetting string value with {@link #resetString()}.
     */
    private char[] origStringChars;

    /**
     * String value contained in the object (for use only if its current type is STRING)
     */
    private RawString stringValue;

    /**
     * Floating point value contained in the object (for use only if its current type is FLOAT)
     */
    private double floatValue;

    /**
     * Date value represented by the number of milliseconds since 1970-01-01, 00:00:00 GMT.
     */
    private long dateValue;

    /**
     * Current type of the value
     */
    private Type type = Type.NONE;

    /**
     * Creates new instance setting maximum length of strings that it can represent using its
     * own storage. Created object can represent longer strings after using storage of some
     * external {@link RawString} object by calling {@link #setString(ReadableRawString)}.
     *
     * @param maxStringLength maximum length of strings that can be contained in the object.
     * @throws IllegalArgumentException in case given <code>maxStringLength</code> is negative
     */
    public Value(int maxStringLength) {
        if (maxStringLength < 0) {
            throw new IllegalArgumentException(
                "Maximum string length must not be negative: " + maxStringLength);
        }
        origStringChars = new char[maxStringLength];
        stringValue = new RawString(origStringChars, 0, 0);
    }

    /**
     * Returns type of current value, indicating which get* method may called in the current
     * state.
     *
     * @return type of current value
     * @see Type
     */
    public Type getType() {
        return type;
    }

    /**
     * Returns integer value contained in the object. The method may be called only if a type
     * of current value is integer, i.e. {@link #getType()} equals <code>Type.INTEGER</code>.
     *
     * @return integer value contained in the object
     */
    public long getInteger() {
        Debugging.assume(
            type == Type.INTEGER,
            "Trying to get integer value from Value object set to type " + type.getName());
        return integerValue;
    }

    /**
     * Sets current value contained in the object to given integer value. After calling this
     * method the object's type will be set to INTEGER ({@link #getType()} will return
     * <code>Type.INTEGER</code>).
     *
     * @param value value to set as current
     */
    public void setInteger(long value) {
        type = Type.INTEGER;
        integerValue = value;
    }

    /**
     * Returns boolean value contained in the object. The method may be called only if a type
     * of current value is boolean, ie {@link #getType()} equals <code>Type.BOOLEAN</code>.
     *
     * @return boolean value contained in the object
     */
    public boolean getBoolean() {
        Debugging.assume(
            type == Type.BOOLEAN,
            "Trying to get boolean value from Value object set to type " + type.getName());
        return booleanValue;
    }

    /**
     * Sets current value contained in the object to given boolean value. After calling this
     * method the object's type will be set to BOOLEAN ({@link #getType()} will return
     * <code>Type.BOOLEAN</code>).
     *
     * @param value value to set as current
     */
    public void setBoolean(boolean value) {
        type = Type.BOOLEAN;
        booleanValue = value;
    }

    /**
     * Returns string value contained in the object. The method may be called only if a type
     * of current value is string, i.e. {@link #getType()} equals <code>Type.STRING</code>.
     *
     * @return string value contained in the object
     */
    public RawString getString() {
        Debugging.assume(
            type == Type.STRING,
            "Trying to get string value from Value object set to type " + type.getName());
        return stringValue;
    }

    /**
     * Sets current value contained in the object to given string value. After calling this
     * method the object's type will be set to STRING ({@link #getType()} will return
     * <code>Type.STRING</code>).
     * <p>
     * The method doesn't modify current storage of underlying {@link RawString} object, it
     * simply assigns storage of given <code>RawString</code> to it (after executing the
     * operation both <code>RawString</code> instances share the same storage).
     * <p>
     * If modification of existing string storage is required, one should access underlying
     * <code>RawString</code> object using {@link #getString()} and use its methods to modify
     * string's characters, e.g.:
     * <p>
     * <pre>
     * value.getString().append("abc");
     * </pre>
     *
     * @param value value to set as current
     */
    public void setString(ReadableRawString value) {
        type = Type.STRING;
        stringValue.set(value);
    }

    /**
     * Sets storage of underlying {@link RawString} object to <code>Value</code>'s own array
     * of characters and sets value of the string to empty string starting from the beginning
     * of the array. After calling this method the object's type will be set to STRING
     * ({@link #getType()} will return <code>Type.STRING</code>).
     */
    public void resetString() {
        type = Type.STRING;
        stringValue.set(origStringChars, 0, 0);
    }

    /**
     * Returns floating point value contained in the object. The method may be called only if a
     * type of current value is floating point number, i.e. {@link #getType()} equals
     * <code>Type.FLOAT</code>.
     *
     * @return floating point value contained in the object
     */
    public double getFloat() {
        Debugging.assume(
            type == Type.FLOAT,
            "Trying to get floating point value from Value object set to type "
            + type.getName());
        return floatValue;
    }

    /**
     * Sets current value contained in the object to given floating point value. After calling
     * this method the object's type will be set to FLOAT ({@link #getType()} will return
     * <code>Type.FLOAT</code>).
     *
     * @param value value to set as current
     */
    public void setFloat(double value) {
        type = Type.FLOAT;
        floatValue = value;
    }

    /**
     * Returns date value contained in the object as number of milliseconds since 1970-01-01,
     * 00:00:00 GMT. The method may be called only if a type of current value is date, ie
     * {@link #getType()} equals <code>Type.DATE</code>.
     *
     * @return date value contained in the object
     */
    public long getDate() {
        Debugging.assume(
            type == Type.DATE,
            "Trying to get date value from Value object set to type "
            + type.getName());
        return dateValue;
    }

    /**
     * Sets current value contained in the object to date value given as number of milliseconds
     * since 1970-01-01, 00:00:00 GMT. After calling this method the object's type will be set
     * to DATE ({@link #getType()} will return <code>Type.DATE</code>).
     *
     * @param value value to set as current
     */
    public void setDate(long value) {
        type = Type.DATE;
        dateValue = value;
    }

    /**
     * Sets value and type of this instance by copying them from the other one. Conceptually, it
     * works like <code>clone()</code> but on existing instance.
     * <p>
     * Note: in case of string values, after calling this method the instance will not be fully
     * independent of the instance it was copied from, because both will be sharing single
     * storage (array of characters) for underlying {@link RawString} instances. Always make
     * sure you know what you're doing when using this method with string values!
     *
     * @param other {@link Value} instance to copy value and type from
     */
    public void set(Value other) {
        Type newType = other.getType();
        switch (newType.getEncodedType()) {
        case Type.NONE_ENCODED:
            break;
        case Type.BOOLEAN_ENCODED:
            booleanValue = other.booleanValue;
            break;
        case Type.INTEGER_ENCODED:
            integerValue = other.integerValue;
            break;
        case Type.STRING_ENCODED:
            stringValue.set(other.stringValue);
            break;
        case Type.FLOAT_ENCODED:
            floatValue = other.floatValue;
            break;
        case Type.DATE_ENCODED:
            dateValue = other.dateValue;
            break;
        default:
            Debugging.assumeNotReached();
        }
        type = newType;
    }

    /**
     * Check whether this instance is equal to another <code>Value</code> instance by comparing
     * their embedded values. Two instances are considered equal if both have the same type and
     * their embedded values are equal.
     *
     * @param other <code>Value</code> instance to compare this one to
     * @return
     *     <code>true</code> if both instances contain equal values, <code>false</code>
     *     otherwise
     */
    public boolean equals(Value other) {
        if (other == null || type != other.type) {
            return false;
        }
        boolean result;
        switch (type.getEncodedType()) {
        case Type.BOOLEAN_ENCODED:
            result = booleanValue == other.booleanValue;
            break;
        case Type.INTEGER_ENCODED:
            result = integerValue == other.integerValue;
            break;
        case Type.STRING_ENCODED:
            result = stringValue.equals(other.stringValue);
            break;
        case Type.FLOAT_ENCODED:
            result = floatValue == other.floatValue;
            break;
        case Type.DATE_ENCODED:
            result = dateValue == other.dateValue;
            break;
        default:
            Debugging.assumeNotReached();
            result = false;
        }
        return result;
    }

    /**
     * Returns string representation of the value (in format appropriate for its type).
     *
     * @return string representation of the value
     */
    public String toString() {
        String str;
        switch (type.getEncodedType()) {
        case Type.NONE_ENCODED:
            str = "NONE";
            break;
        case Type.BOOLEAN_ENCODED:
            if (booleanValue) {
                str = "true";
            } else {
                str = "false";
            }
            break;
        case Type.INTEGER_ENCODED:
            str = Long.toString(integerValue);
            break;
        case Type.STRING_ENCODED:
            str = stringValue.toString();
            break;
        case Type.FLOAT_ENCODED:
            str = Double.toString(floatValue);
            break;
        case Type.DATE_ENCODED:
            Date date = new Date(dateValue);
            SimpleDateFormat dateFormatter = new SimpleDateFormat(DATE_FORMAT);
            dateFormatter.setCalendar(new SimpleCalendar());
            str = dateFormatter.format(date);
            break;
        default:
            Debugging.assumeNotReached();
            str = null;
        }
        return str;
    }

    /**
     * Checks whether this <code>Value</code> instance is equal to another object.
     *
     * @param other object to compare this instance to
     * @return
     *     <code>true</code> if <code>other</code> is <code>Value</code> instance with value
     *     equal to this one, <code>false</code> otherwise
     */
    public boolean equals(Object other) {
        if (other instanceof Value) {
            return equals((Value) other);
        } else {
            return false;
        }
    }

    /**
     * Returns a hash code value for this instance, using embedded value for the calculation.
     *
     * @return returns a hash code value for this instance
     */
    public int hashCode() {
        int hash;
        switch (type.getEncodedType()) {
        case Type.BOOLEAN_ENCODED:
            hash = booleanValue ? 1 : 0;
            break;
        case Type.INTEGER_ENCODED:
            hash = LongUtils.hashCode(integerValue);
            break;
        case Type.STRING_ENCODED:
            hash = stringValue.hashCode();
            break;
        case Type.FLOAT_ENCODED:
            hash = LongUtils.hashCode(Double.doubleToRawLongBits(floatValue));
            break;
        case Type.DATE_ENCODED:
            hash = LongUtils.hashCode(dateValue);
            break;
        default:
            Debugging.assumeNotReached();
            hash = 0;
        }
        return hash;
    }

}
