/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.core.functions;

import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.RawString;


/**
 * Base class for functions generating results of string type.
 * Type of result is represented by value Type.STRING.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class StringFunction extends Function {

    /**
     * Evaluates the function's value for currently set parameters by setting the result to
     * empty string and then calling {@link #execute(RawString)} method.
     *
     * @param result
     *     {@link Value} object to store function's result in; its type will be set to integer
     *     (<code>Type.STRING</code>)
     * @see Function#typedEvaluate(Value)
     */
    protected final void typedEvaluate(Value result) {
        result.resetString();
        execute(result.getString());
    }

    /**
     * Executes function's logic and evaluates result. This method returns the result by
     * appending data to string value contained in <code>result</code> object.<p>
     * The preferred method for building the result is using one of
     * <code>RawString.append</code> methods. Initially the result is set to empty string, so
     * the method doesn't need to do anything if empty string is the expected result.
     * <p>
     * Note: {@link pl.chyla.wyjec.utils.RawString.RawStringOverflowException} exceptions thrown
     * by <code>RawString.append</code> methods should not be caught as they're handled inside
     * Wyjec core.
     *
     * @param result object containing evaluation result, to be updated by the method
     */
    protected abstract void execute(RawString result);

    /**
     * @return type of result of the function; always returns Type.STRING
     * @see Function#getResultType()
     */
    public final Type getResultType() {
        return Type.STRING;
    }

}
