/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.evaluator;

import java.util.Arrays;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.IEvaluationContext;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.parser.AtomNode;
import pl.chyla.wyjec.parser.AtomType;
import pl.chyla.wyjec.parser.ExpressionTree;
import pl.chyla.wyjec.parser.FunctionNode;
import pl.chyla.wyjec.parser.ParameterNode;
import pl.chyla.wyjec.utils.IObjectsFactory;
import pl.chyla.wyjec.utils.ObjectsPool;


/**
 * Evaluator for Wyjec expression language. Evaluation is based on expression trees produced by
 * Wyjec parser and the result is always single {@link Value} instance.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Evaluator {

    /**
     * Pool of {@link Value} objects for use as temporary values (actual parameters and results
     * of functions) during evaluation.
     */
    private final ObjectsPool valuePool;

    /**
     * Pool of {@link DirectEvaluationContext} objects for passing parameters to functions and
     * getting results when not using lazy evaluation.
     */
    private final ObjectsPool directEvalContextPool;

    /**
     * Pool of {@link DirectEvaluationContext} objects for passing parameters to functions and
     * getting results when evaluating parameters lazily.
     */
    private final ObjectsPool lazyEvalContextPool;

    /**
     * Constructs new evaluator for Wyjec expressions using given configuration.
     *
     * @param config
     *     object defining various configuration details of the evaluator (mainly limits of
     *     resources)
     */
    public Evaluator(Config config) {
        final int maxStringLength = config.getStringLengthLimit();
        valuePool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new Value(maxStringLength);
                }
            }, config.getTmpValuesCountSoftLimit());
        final int paramsSoftLimit = config.getFuncActualParamsSoftLimit();
        directEvalContextPool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new DirectEvaluationContext(paramsSoftLimit);
                }
            }, config.getFuncDepthSoftLimit());
        lazyEvalContextPool = new ObjectsPool(
            new IObjectsFactory() {
                public Object createObject() {
                    return new LazyEvaluationContext(paramsSoftLimit);
                }
            }, config.getFuncDepthSoftLimit());
    }

    /**
     * Evaluates given expression tree to single {@link Value} instance. During evaluation
     * internal pool of objects is used for storing temporary values, so the method doesn't
     * create or destroy any objects unless complexity of the expression exceeds limits defined
     * in configuration passed to the evaluator during its construction.
     * <p>
     * Note: returned {@link Value} instance will be reused by the evaluator next time this
     * method is called, so it will be invalid and may not be used after that. Also be careful
     * when copying the result using {@link Value#set(Value)} method, because the copy prepared
     * this way will not be fully independent from the evaluator.
     *
     * @param exprTree expression tree to evaluate
     * @return value the result of evaluation
     * @throws EvaluationException in case of error during expression evaluation.
     */
    public Value evaluate(ExpressionTree exprTree) throws EvaluationException {
        Value result = (Value) valuePool.getObject();
        try {
            evaluateAtom(exprTree.getTopLevelAtom(), result);
        } catch (FunctionEvaluationException e) {
            throw new EvaluationException(e);
        } finally {
            valuePool.reset();
            directEvalContextPool.reset();
            lazyEvalContextPool.reset();
        }
        return result;
    }

    /**
     * Evaluates given atom to single {@link Value} instance got from the internal pool.
     *
     * @param atom atom to evaluate
     * @param result {@link Value} instance to store the result in
     */
    private void evaluateAtom(AtomNode atom, Value result) {
        if (atom.getAtomType() == AtomType.VALUE) {
            result.set(atom.getAsValueNode().getValue());
        } else {
            evaluateFunction(atom.getAsFunctionNode(), result);
        }
    }

    /**
     * Evaluates given function atom by evaluating function embedded in atom description using
     * accompanying parameter nodes for function's parameters. If the function doesn't support
     * lazy evaluation of its parameters, first all parameters are evaluated and then
     * {@link DirectEvaluationContext} instance is used to evaluate the function itself. If the
     * function supports lazy evaluation of parameters, function's evaluation starts
     * immediately using {@link LazyEvaluationContext} and all parameters are later evaluated
     * on demand.
     *
     * @param funcAtom atom representing reference to function to evaluate
     * @param result {@link Value} instance to store the result in
     */
    private void evaluateFunction(FunctionNode funcAtom, Value result) {
        Function func = funcAtom.getFunction();
        ObjectsPool evalContextPool;
        if (func.getSignature().getLazyEvaluationSupport()) {
            evalContextPool = lazyEvalContextPool;
        } else {
            evalContextPool = directEvalContextPool;
        }
        BaseEvaluationContext evalContext = (BaseEvaluationContext) evalContextPool.getObject();
        evalContext.prepare(funcAtom.getParameters(), result);
        try {
            func.evaluate(evalContext);
        } catch (Exception e) {
            throw new FunctionEvaluationException(func, e);
        }
        evalContext.releaseParameters();
        evalContextPool.releaseObject(evalContext);
    }


    /**
     * Base class for Evaluator's implementations of {@link IEvaluationContext} interface. It
     * provides access to {@link Value} instance for function's result and defines two abstract
     * methods related to handling of function's parameters.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private abstract class BaseEvaluationContext implements IEvaluationContext {

        /**
         * {@link Value} instance holding the result of function evaluation
         */
        private Value result;

        /**
         * Number of actual parameters for current function evaluation
         */
        private int paramsCount;

        /**
         * Initializes this instance for function evaluation, using given parameter nodes and
         * object for storing function's result.
         *
         * @param funcParamsNode
         *     expression tree node representing the first parameter of the function (and
         *     referencing other parameters)
         * @param funcResult
         *     {@link Value} instance to store the result of function's evaluation in
         */
        final void prepare(ParameterNode funcParamsNode, Value funcResult) {
            result = funcResult;
            int count = 0;
            for (ParameterNode param = funcParamsNode; param != null; param = param.getNext()) {
                count++;
            }
            paramsCount = count;
            prepareParameters(funcParamsNode);
        }

        /**
         * Prepares actual parameters for function's evaluation, using information stored in
         * given parameter nodes from expression tree.
         *
         * @param funcParamsNode
         *     expression tree node representing the first parameter of the function (and
         *     referencing other parameters)
         */
        abstract void prepareParameters(ParameterNode funcParamsNode);

        /**
         * Releases all {@link Value} instances got from the pool for storing values of
         * function's parameters, i.e. returns them back to the pool.
         */
        abstract void releaseParameters();

        /**
         * {@inheritDoc}
         *
         * @return {@inheritDoc}
         */
        public final Value getResult() {
            return result;
        }

        /**
         * {@inheritDoc}
         */
        public final int getParametersCount() {
            return paramsCount;
        }

    }


    /**
     * Simple implementation of {@link IEvaluationContext} interface, that provides function's
     * parameters directly from internal array of {@link Value}s.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private final class DirectEvaluationContext extends BaseEvaluationContext {

        /**
         * Preallocated array for storing values of function's parameters. Used as
         * {@link #paramsValues} when possible.
         */
        private final Value[] stdParamsValues;

        /**
         * Values of function's parameters (for direct use).
         */
        private Value[] paramsValues;

        /**
         * Constructs <code>DirectEvaluationContext</code> instance.
         *
         * @param paramsSoftLimit soft limit for a number of actual parameters
         */
        DirectEvaluationContext(int paramsSoftLimit) {
            stdParamsValues = new Value[paramsSoftLimit];
        }

        /**
         * Initializes actual parameters for function's evaluation by evaluating atoms contained
         * in given parameter nodes from expression tree. Values of actual parameters will be
         * stored in {@link #stdParamsValues} array.
         *
         * @param funcParamsNode {@inheritDoc}
         */
        void prepareParameters(ParameterNode funcParamsNode) {
            int paramsCount = getParametersCount();
            if (paramsCount <= stdParamsValues.length) {
                paramsValues = stdParamsValues;
            } else {
                paramsValues = new Value[paramsCount];
            }
            int paramNo = 0;
            for (ParameterNode param = funcParamsNode; param != null; param = param.getNext()) {
                Value paramValue = (Value) valuePool.getObject();
                evaluateAtom(param.getAtom(), paramValue);
                paramsValues[paramNo] = paramValue;
                paramNo++;
            }
        }

        /**
         * {@inheritDoc}
         */
        void releaseParameters() {
            int paramsCount = getParametersCount();
            for (int i = 0; i < paramsCount; i++) {
                valuePool.releaseObject(paramsValues[i]);
            }
        }

        /**
         * Returns {@link Value} instance with value of requested actual parameter by getting
         * it directly from previously prepared array of parameters.
         *
         * @param paramIndex index of actual parameter to get
         * @return function's actual parameter with given index
         */
        public Value getParameter(int paramIndex) {
            return paramsValues[paramIndex];
        }

    }


    /**
     * Implementation of {@link IEvaluationContext} interface, that uses lazy approach for
     * evaluation of function's actual parameters.
     *
     * @author Zbigniew Chyla (zbigniew@chyla.pl)
     */
    private final class LazyEvaluationContext extends BaseEvaluationContext {

        /**
         * Preallocated array for storing values of function's parameters. Used as
         * {@link #paramsValues} when possible.
         */
        private final Value[] stdParamsValues;

        /**
         * Values of function's parameters, filled with <code>null</code>s before function's
         * evaluation and updated later with real values of parameters when they're requested
         * requested by the function.
         */
        private Value[] paramsValues;

        /**
         * Preallocated array for storing atoms describing function's parameters. Used as
         * {@link #paramsAtoms} when possible.
         */
        private final AtomNode[] stdParamsAtoms;

        /**
         * Atoms describing function's parameters, copied from expression tree.
         */
        private AtomNode[] paramsAtoms;

        /**
         * Constructs <code>LazyEvaluationContext</code> instance.
         *
         * @param paramsSoftLimit soft limit for a number of actual parameters
         */
        LazyEvaluationContext(int paramsSoftLimit) {
            stdParamsValues = new Value[paramsSoftLimit];
            stdParamsAtoms = new AtomNode[paramsSoftLimit];
        }

        /**
         * Initializes actual parameters for function's evaluation by copying atom nodes from
         * given parameter nodes to {@link #paramsAtoms} array. The atoms will not be evaluated
         * immediately; instead, the array of parameters' values ({@link #paramsValues}) will
         * be initialized with <code>null</code>s and the real values will be evaluated on
         * demand using corresponding elements from {@link #paramsAtoms} array.
         *
         * @param funcParamsNode {@inheritDoc}
         */
        void prepareParameters(ParameterNode funcParamsNode) {
            int paramsCount = getParametersCount();
            if (paramsCount <= stdParamsValues.length) {
                paramsValues = stdParamsValues;
            } else {
                paramsValues = new Value[paramsCount];
            }
            if (paramsCount <= stdParamsAtoms.length) {
                paramsAtoms = stdParamsAtoms;
            } else {
                paramsAtoms = new AtomNode[paramsCount];
            }
            Arrays.fill(paramsValues, 0, paramsCount, null);
            int paramNo = 0;
            for (ParameterNode param = funcParamsNode; param != null; param = param.getNext()) {
                paramsAtoms[paramNo] = param.getAtom();
                paramNo++;
            }
        }

        /**
         * {@inheritDoc}
         */
        void releaseParameters() {
            int paramsCount = getParametersCount();
            for (int i = 0; i < paramsCount; i++) {
                Value paramValue = paramsValues[i];
                if (paramValue != null) {
                    valuePool.releaseObject(paramValue);
                }
            }
        }

        /**
         * Returns {@link Value} instance with value of requested actual parameter by getting
         * it directly from {@link #paramsValues} array if it's available or by evaluating the
         * corresponding element from {@link #paramsAtoms} array and then storing the result
         * in {@link #paramsValues}. This approach is called lazy evaluation.
         *
         * @param paramIndex index of actual parameter to get
         * @return function's actual parameter with given index
         */
        public Value getParameter(int paramIndex) {
            Value paramValue = paramsValues[paramIndex];
            if (paramValue == null) {
                paramValue = (Value) valuePool.getObject();
                evaluateAtom(paramsAtoms[paramIndex], paramValue);
                paramsValues[paramIndex] = paramValue;
            }
            return paramValue;
        }

    }

}
