/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.extfuncs.langfuncs;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;


/**
 * Uses Polish grammar rules to choose the appropriate plural form for number specified as the
 * first parameter and returns one of the remaining three string parameters.
 *
 * @wyjec.example "{pluralPL|{3}|czapka|czapki|czapek}" "czapki" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public abstract class FnBasePluralPL extends StringFunction {

    /**
     * Number determining plural form to use
     * @wyjec.param number
     */
    private static final IntegerParameter pNumber = new IntegerParameter();

    /**
     * Singular form of a word (used when number equals 1)
     * @wyjec.param singular
     */
    private static final StringParameter pSingular = new StringParameter();

    /**
     * First plural form of a word (used e.g. when number equals 2)
     * @wyjec.param plural1
     */
    private static final StringParameter pPlural1 = new StringParameter();

    /**
     * Second plural form of a word (used e.g. when number equals 5)
     * @wyjec.param plural2
     */
    private static final StringParameter pPlural2 = new StringParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pNumber, pSingular, pPlural1, pPlural2}, null);

    public final FunctionSignature getSignature() {
        return signature;
    }

    protected final void execute(RawString result) {
        long number = getParameter(pNumber);
        if (number == 1) {
            result.append(getParameter(pSingular));
        } else {
            long rem10 = number % 10;
            long rem100 = number % 100;
            if (rem10 >= 2 && rem10 <= 4 && (rem100 < 10 || rem100 >= 20)) {
                result.append(getParameter(pPlural1));
            } else {
                result.append(getParameter(pPlural2));
            }
        }
    }

}
