/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.lexer;

import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Lexical analyser for Simple Expression Language. Instance of this class may be used many
 * times for analysing different input strings, at the beginning of each analysis
 * {@link #start(ReadableRawString)} method should be called.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Lexer {

    /**
     * Character representing CHAR_FUNCTION_START token
     */
    private static final char CHAR_FUNCTION_START = '{';

    /**
     * Character representing CHAR_SEPARATOR token
     */
    private static final char CHAR_SEPARATOR = '|';

    /**
     * Character representing CHAR_FUNCTION_END token
     */
    private static final char CHAR_FUNCTION_END = '}';

    /**
     * String currently analysed by this lexer (set after every call to
     * {@link #start(ReadableRawString)})
     */
    private ReadableRawString input;

    /**
     * Copy of <code>input.getStorage()</code> for faster access
     * (set inside {@link #start(ReadableRawString)}).
     */
    private char[] inputChars;

    /**
     * Index of next character to be processed from <code>inputChars</code> array
     */
    private int inputPos;

    /**
     * Copy of <code>newInput.getEndPos()</code> for faster access
     * (set inside {@link #start(ReadableRawString)}).
     */
    private int inputPosLimit;

    /**
     * Index of first character of the current token in {@link #inputChars} array
     */
    private int tokenPos;

    /**
     * String containing value of current token (with characters from {@link #inputChars}) as
     * <code>RawString</code>'s storage. Value is updated on demand in {@link #getTokenString()}
     * method.
     */
    private RawString tokenString = new RawString(0);

    /**
     * <code>int</code> representation of the current token
     */
    private int tokenType = Token.TERMINATOR;

    /**
     * Constructs <code>Lexer</code> instance setting it as if was reading from empty string,
     * i.e. current token will be <code>Token.TERMINATOR</code>.
     */
    public Lexer() {
        start(new RawString(0));
    }

    /**
     * Resets state of the lexer and starts analysing of new input string. After calling this
     * method the first token from input string will be set as current token for the lexer (it
     * will be set to {@link Token#TERMINATOR} if the input is empty).
     *
     * @param newInput input string to start analysing
     */
    public void start(ReadableRawString newInput) {
        input = newInput;
        inputChars = newInput.getStorage();
        inputPos = newInput.getStartPos();
        inputPosLimit = newInput.getEndPos();
        tokenPos = inputPos;
        tokenString.set(newInput.getStorage(), tokenPos, 0);
        getNextToken();
    }

    /**
     * Reads next token from input string and updates the state of the lexer accordingly. In
     * case there are no characters on input left, current token will be set to
     * {@link Token#TERMINATOR}.
     *
     * @return <code>int</code> representation of the new current token
     */
    public int getNextToken() {
        if (Debugging.DEBUG_ENABLED) {
            Debugging.assume(
                inputPos <= inputPosLimit,
                "trying to read beyond the end of input string");
        }
        tokenPos = inputPos;
        if (inputPos == inputPosLimit) {
            tokenType = Token.TERMINATOR;
            return tokenType;
        }
        char[] chars = inputChars;
        switch (chars[inputPos]) {
        case CHAR_FUNCTION_START:
            inputPos++;
            tokenType = Token.FUNCTION_START;
            break;
        case CHAR_FUNCTION_END:
            inputPos++;
            tokenType = Token.FUNCTION_END;
            break;
        case CHAR_SEPARATOR:
            inputPos++;
            tokenType = Token.SEPARATOR;
            break;
        default:
            do {
                inputPos++;
            } while (inputPos < inputPosLimit && chars[inputPos] != CHAR_FUNCTION_START
                     && chars[inputPos] != CHAR_FUNCTION_END
                     && chars[inputPos] != CHAR_SEPARATOR);
            tokenType = Token.STRING;
            break;
        }
        return tokenType;
    }

    /**
     * Returns <code>int</code> representation of the current token (it will return
     * {@link Token#TERMINATOR} if the lexer is at the end of input).
     *
     * @return <code>int</code> representation of the current token
     */
    public int getTokenType() {
        return tokenType;
    }

    /**
     * Returns string containing characters from the input representing value of the current
     * token. Returned string will be sharing storage with string used as input of the lexer
     * (as returned by {@link #getInputString()}).
     *
     * @return
     *     string representing value of the current token, it should not be modified; if the
     *     current token is Token#TERMINATOR, the returned string will be empty
     */
    public ReadableRawString getTokenString() {
        tokenString.setRange(tokenPos, inputPos - tokenPos);
        return tokenString;
    }

    /**
     * Returns index of first character of the current token in the input string.
     *
     * @return
     *     index of first character of the current token in the input string; for
     *     {@link Token#TERMINATOR} token the length of the input string will be returned
     */
    public int getTokenStartIndex() {
        return tokenPos - input.getStartPos();
    }

    /**
     * Returns current input of the lexer.
     *
     * @return
     *     string currently processed by the lexer (empty string if no input has been set using
     *     {@link #start(ReadableRawString)}), it should not be modified
     */
    public ReadableRawString getInputString() {
        return input;
    }

}
