/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.parser;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.lexer.Lexer;
import pl.chyla.wyjec.lexer.Token;
import pl.chyla.wyjec.repo.FunctionsRepository;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Parser and semantic analyser for Wyjec expression language. It's responsible for building
 * expression tree for any Wyjec expression stored in given {@link ReadableRawString} object.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class Parser {

    /**
     * Maximum number of parameters which types are independent of each other. For longer lists
     * of parameters all parameters with indices greater than or equal to this value must have
     * the same type as the one at <code>INDEPENDENT_PARAMS_TYPES_LIMIT - 1</code>.
     */
    private static final int INDEPENDENT_PARAMS_TYPES_LIMIT =
        Config.FUNC_REQ_PARAMS_COUNT_LIMIT + 1;

    /**
     * Atom representing very special constant value: empty string.
     */
    private static final ValueNode EMPTY_STRING_VALUE_ATOM = new ValueNode();
    static {
        Value emptyString = new Value(0);
        emptyString.resetString();
        EMPTY_STRING_VALUE_ATOM.set(emptyString);
    }

    /**
     * Repository of functions used during semantic analysis of expression and building
     * expression tree.
     */
    private final FunctionsRepository funcsRepo;

    /**
     * Constants parser used during semantic analysis of expression and building expression
     * tree.
     */
    private final IConstantsParser constsParser;

    /**
     * Instance of Wyjec lexer that produces stream of tokens for the parser.
     */
    private final Lexer lexer;

    /**
     * Object used to get references to some special functions (conversion functions,
     * concatenation functions) based on information from {@link Config} instance passed to the
     * parser.
     */
    private final SpecialFunctionsProvider specialFuncsProv;

    /**
     * Pool used as the source of objects forming expression tree build by this parser. These
     * objects are nodes of the expression tree.
     */
    private final ExprTreePool exprTreePool;

    /**
     * Temporary value to avoid dynamic allocation of objects: array of types of actual
     * parameters for currently analysed function.
     */
    private final Type[] tmpParamsTypes;

    /**
     * Temporary value to avoid dynamic allocation of objects: name of currently analysed
     * function.
     */
    private final RawString tmpFuncName;

    /**
     * Constructs <code>Parser</code> instance with its own lexer.
     *
     * @param config
     *     configuration to use during parsing and analysing of expressions by this parser
     * @param funcs array of functions to use during semantic analysis of expressions
     * @param constsParser constants parser to use during semantic analysis of expressions
     * @throws IllegalArgumentException
     *     in case any of the constructor's parameters are set to <code>null</code>
     */
    public Parser(Config config, Function[] funcs, IConstantsParser constsParser) {
        if (config == null || funcs == null || constsParser == null) {
            throw new IllegalArgumentException(
                "Neither of the constructor's parameters may be null.");
        }
        funcsRepo = new FunctionsRepository(funcs);
        this.constsParser = constsParser;
        lexer = new Lexer();
        specialFuncsProv = new SpecialFunctionsProvider(config, funcsRepo);
        exprTreePool = new ExprTreePool(1, config);
        tmpParamsTypes = new Type[INDEPENDENT_PARAMS_TYPES_LIMIT];
        tmpFuncName = new RawString(0);
    }

    /**
     * Parses expression stored in given string and builds the corresponding expression tree.
     * <p>
     * Note: generated expression tree will be build from objects got from the parser's internal
     * objects pool. These objects will be reused every time this method is called, so returned
     * expression tree will be valid only until the next call of this method. However, it's
     * possible to get persistent copy of expression tree (independent from the parser) by
     * calling its {@link ExpressionTree#makeDeepCopy(ExprTreePool)} method.
     *
     * @param input string containing expression to parse
     * @return expression tree corresponding to expression contained in the string
     * @throws InvalidInputException
     *    in case of syntax error in given input or some semantic error, including: referencing
     *    to undefined function or constant, missing string conversion function for type used as
     *    concatenation element.
     */
    public ExpressionTree parse(ReadableRawString input) throws InvalidInputException {
        lexer.start(input);
        ExpressionTree exprTree;
        try {
            AtomNode topLevelAtom = parseAtomSequence();
            checkTokenType(Token.TERMINATOR);
            exprTree = exprTreePool.getExpressionTree();
            exprTree.set(topLevelAtom);
        } finally {
            exprTreePool.reset();
        }
        return exprTree;
    }

    /**
     * Parses a sequence of atoms ("atom_sequence" nonterminal in Wyjec grammar) and returns
     * atom representing reference to string concatenation function with all sequence elements
     * as its parameters. This method uses {@link SpecialFunctionsProvider} to get an instance
     * of the appropriate string concatenation function.
     * <p>
     * Precondition: current token is the first token of atom sequence<br>
     * Postcondition: current token is a token after the last atom of atom sequence
     *
     * @return
     *     atom representing reference to string concatenation function with all parsed sequence
     *     elements as its parameters
     * @throws InvalidInputException in case of syntax error
     */
    private AtomNode parseAtomSequence() throws InvalidInputException {
        int sequenceStartPos = lexer.getTokenStartIndex();
        AtomNode firstAtom = parseAtom();
        if (firstAtom == null) {
            firstAtom = EMPTY_STRING_VALUE_ATOM;
        }
        if (lexer.getTokenType() != Token.FUNCTION_START
            && lexer.getTokenType() != Token.STRING) {
            return firstAtom;
        }
        int atomsCount = 1;
        ParameterNode firstParam = exprTreePool.getParameterNode();
        firstParam.setFirst(assureStringAtom(firstAtom, sequenceStartPos));
        ParameterNode lastParam = firstParam;
        while (true) {
            int paramAtomStartPos = lexer.getTokenStartIndex();
            AtomNode paramAtom = parseAtom();
            if (paramAtom == null) {
                break;
            }
            ParameterNode param = exprTreePool.getParameterNode();
            param.setAfter(lastParam, assureStringAtom(paramAtom, paramAtomStartPos));
            lastParam = param;
            atomsCount++;
        }
        lastParam.setLast();
        FunctionNode funcAtom = exprTreePool.getFunctionNode();
        Function concatFunction = specialFuncsProv.getStrConcatenationFunc(atomsCount);
        if (concatFunction == null) {
            throw new InvalidInputException(
                lexer, sequenceStartPos,
                "There's no string concatenation function with " + atomsCount + "parameters.");
        }
        funcAtom.set(concatFunction, firstParam);
        return funcAtom;
    }

    /**
     * Parses single atom and generates an {@link AtomNode} instance that represents it.
     * <p>
     * Precondition: current token is the first token of the atom to parse or any token in case
     * there's no atom to parse<br>
     * Postcondition: current token is a token after the atom or unchanged token in case there's
     * no atom to parse
     *
     * @return
     *     generated {@link AtomNode} instance or <code>null</code> if there's no recognized
     *     atom at the current position
     * @throws InvalidInputException in case of syntax error during parsing the atom
     */
    AtomNode parseAtom() throws InvalidInputException {
        AtomNode atom;
        if (lexer.getTokenType() == Token.FUNCTION_START) {
            atom = parseFunction();
        } else if (lexer.getTokenType() == Token.STRING) {
            atom = parseLiteral();
        } else {
            atom = null;
        }
        return atom;
    }

    /**
     * Checks a type of given atom and in case it's not {@link Type#STRING}, wraps the atom with
     * conversion function to assure that returned atom has string type. The method will get
     * string conversion function appropriate for type of given atom, using
     * {@link SpecialFunctionsProvider}.
     *
     * @param atom atom to check type for and optionally to wrap with string conversion function
     * @param atomStartPos
     *     position of first token of the atom in the input stream, used to as part of error
     *     message in case of throwing {@link InvalidInputException}
     * @return
     *     unmodified <code>atom</code> instance if its type is {@link Type#STRING} or another
     *     atom used as wrapper around <code>atom</code>, responsible for converting its value
     *     to string
     * @throws InvalidInputException
     *     in case given's atom type is not {@link Type#STRING} and there's no conversion
     *     function for its type
     */
    private AtomNode assureStringAtom(AtomNode atom, int atomStartPos)
        throws InvalidInputException {
        if (atom.getValueType() == Type.STRING) {
            return atom;
        }
        Function convFunction = specialFuncsProv.getStrConversionFunc(atom.getValueType());
        if (convFunction == null) {
            throw new InvalidInputException(
                lexer, atomStartPos,
                "Value is used as concatenation element, but there's no conversion function for"
                + " values of this type (" + atom.getValueType().getName() + ")");
        }
        ParameterNode param = exprTreePool.getParameterNode();
        param.setFirst(atom);
        param.setLast();
        FunctionNode funcAtom = exprTreePool.getFunctionNode();
        funcAtom.set(convFunction, param);
        return funcAtom;
    }

    /**
     * Parses part of the input that describes reference to a function (second alternative in
     * production describing "atom" nonterminal in Wyjec grammar) and returns atom representing
     * this reference (with resolved function instance and instances of atoms representing
     * function parameters).
     * <p>
     * Precondition: current token is {@link Token#FUNCTION_START}<br>
     * Postcondition: current token is a token after {@link Token#FUNCTION_END} corresponding to
     * {@link Token#FUNCTION_START} that was current on this method's entry
     *
     * @return atom representing parsed reference to function
     * @throws InvalidInputException
     *     in case of syntax error or inability to find specified function (with given name and
     *     parameters' types) in the function repository
     */
    private AtomNode parseFunction() throws InvalidInputException {
        lexer.getNextToken();
        checkTokenType(Token.STRING);
        int funcNamePos = lexer.getTokenStartIndex();
        ReadableRawString funcName = lexer.getTokenString();
        char[] funcNameStorage = funcName.getStorage();
        int funcNameStartPos = funcName.getStartPos();
        int funcNameLength = funcName.getLength();
        lexer.getNextToken();
        ParameterNode params = parseParameterSequence();
        checkTokenType(Token.FUNCTION_END);
        lexer.getNextToken();
        tmpFuncName.set(funcNameStorage, funcNameStartPos, funcNameLength);
        return makeFunctionAtom(tmpFuncName, funcNamePos, params);
    }

    /**
     * Searches the function repository for function with given name that can accept parameters
     * described by given parameter nodes and returns atom representing reference to this
     * function with given parameter nodes.
     *
     * @param name name of function to search
     * @param namePos position of the function's name in the input (for error reporting)
     * @param params
     *     nodes representing parameters of the function or <code>null</code> if there're no
     *     parameters for the function
     * @return atom representing reference to function matching given name and parameters' types
     * @throws InvalidInputException
     *     in case the function repository doesn't contain any function with given name that
     *     can accept given parameters
     */
    private AtomNode makeFunctionAtom(ReadableRawString name, int namePos, ParameterNode params)
        throws InvalidInputException {
        int paramsTypesCount = 0;
        ParameterNode param = params;
        while (param != null && paramsTypesCount < INDEPENDENT_PARAMS_TYPES_LIMIT) {
            tmpParamsTypes[paramsTypesCount] = param.getAtom().getValueType();
            paramsTypesCount++;
            param = param.getNext();
        }
        if (paramsTypesCount == INDEPENDENT_PARAMS_TYPES_LIMIT) {
            Type lastType = tmpParamsTypes[INDEPENDENT_PARAMS_TYPES_LIMIT - 1];
            while (param != null) {
                if (param.getAtom().getValueType() != lastType) {
                    throw newMissingFunctionException(name, namePos, params);
                }
                param = param.getNext();
            }
        }
        AtomNode atom;
        Function func = funcsRepo.lookup(tmpFuncName, tmpParamsTypes, paramsTypesCount);
        if (func != null) {
            FunctionNode funcAtom = exprTreePool.getFunctionNode();
            funcAtom.set(func, params);
            atom = funcAtom;
        } else if (paramsTypesCount == 0) {
            Value constValue = exprTreePool.getValue();
            if (!constsParser.parseConstant(constValue, tmpFuncName)) {
                throw newMissingFunctionException(name, namePos, params);
            }
            ValueNode valueAtom = exprTreePool.getValueNode();
            valueAtom.set(constValue);
            atom = valueAtom;
        } else {
            throw newMissingFunctionException(name, namePos, params);
        }
        return atom;
    }

    /**
     * Creates new instance of {@link InvalidInputException} that indicates missing Wyjec
     * function with given name that can accept given actual parameters.
     *
     * @param funcName the function's name
     * @param funcNamePos position of the function's name in the input (for error reporting)
     * @param funcParams
     *     nodes representing parameters of the function or <code>null</code> if there're no
     *     parameters for the function
     * @return new {@link InvalidInputException} instance
     */
    private InvalidInputException newMissingFunctionException(
        ReadableRawString funcName, int funcNamePos, ParameterNode funcParams) {
        StringBuffer msg = new StringBuffer();
        msg.append(
            "There's no function \"" + funcName.toString() + "\""
            + " that accepts the parameters list: ");
        msg.append('(');
        for (ParameterNode param = funcParams; param != null; param = param.getNext()) {
            if (param != funcParams) {
                msg.append(',');
            }
            msg.append(param.getAtom().getValueType().getName());
        }
        msg.append(')');
        return new InvalidInputException(lexer, funcNamePos, msg.toString());
    }

    /**
     * Parses single string literal from the input and returns atom representing its (string)
     * value.
     * <p>
     * Precondition: current token is {@link Token#STRING}<br>
     * Postcondition: current token is a token after {@link Token#STRING} that was current on
     * this method's entry
     *
     * @return atom representing constant value of parsed literal
     */
    private AtomNode parseLiteral() {
        ValueNode valueAtom = exprTreePool.getValueNode();
        Value value = exprTreePool.getValue();
        value.setString(lexer.getTokenString());
        valueAtom.set(value);
        lexer.getNextToken();
        return valueAtom;
    }

    /**
     * Parses a sequence of function parameters, described in Wyjec grammar using
     * "parameter_list" nonterminal, and returns parameter node representing the first
     * parameter and connected to following parameters. If there're no parameters present (the
     * grammar permits this), the method returns <code>null</code>.
     * <p>
     * Precondition: current token is the first token after the function's name<br>
     * Postcondition: current token is a token after the last parameter
     *
     * @return
     *     parameter node representing the first parameter or <code>null</code> if no parameters
     *     are present
     * @throws InvalidInputException in case of syntax error
     */
    private ParameterNode parseParameterSequence() throws InvalidInputException {
        if (lexer.getTokenType() != Token.SEPARATOR) {
            return null;
        }
        lexer.getNextToken();
        ParameterNode firstParam = exprTreePool.getParameterNode();
        firstParam.setFirst(parseAtomSequence());
        ParameterNode lastParam = firstParam;
        while (lexer.getTokenType() == Token.SEPARATOR) {
            lexer.getNextToken();
            ParameterNode param = exprTreePool.getParameterNode();
            param.setAfter(lastParam, parseAtomSequence());
            lastParam = param;
        }
        lastParam.setLast();
        return firstParam;
    }

    /**
     * Checks whether type of current token is equal to one given as a parameter.
     *
     * @param expectedTokenType
     *     <code>int</code> representation of token type to check current token against
     * @throws InvalidInputException
     *     in case the type of current token is not <code>expectedTokenType</code>
     */
    private void checkTokenType(int expectedTokenType) throws InvalidInputException {
        if (lexer.getTokenType() != expectedTokenType) {
            throw new InvalidInputException(
                lexer,
                "Unexpected token " + Token.getName(lexer.getTokenType())
                + ": \"" + lexer.getTokenString().toString() + "\"");
        }
    }

}
