/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.utils.IntUtils;


/**
 * Helper class responsible for encoding a sequence of {@link Type} instances into single
 * <code>long</code> value and containing some operations on such encoded representations.
 * Additionally, encoded form may contain information about "extension" of the sequence, i.e.
 * single type added at the end of sequence multiple times such that total number of types is
 * equal to defined maximum.
 * <p>
 * Every number encoded using the class may represent one of two different entities:
 * <ul>
 * <li>
 * Sequence of types of simple values, e.g. sequence of types of actual parameters passed to a
 * function. In this case there's no extension - {@link Type#NONE} should be used as an
 * extension.
 * </li>
 * <li>
 * Sequence of types that don't correspond to any concrete values and are used as declaration
 * of possible sequences of types of values, e.g. sequence of types of declared parameters of a
 * function. In this case types from a sequence represent types of required values (required
 * actual parameters of a function) and an extension represents a type of optional values (any
 * number of them) after the required ones. If only required types are accepted,
 * {@link Type#NONE} should be used as an extension (it means that no extension exists).
 * </li>
 * </ul>
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
final class EncodedTypeList {

    /**
     * Maximum length of sequence of types that can be encoded using our scheme
     */
    private static final int MAX_ENCODED_TYPE_LIST_LENGTH =
        IntUtils.INT_BITS_COUNT / Type.ENCODED_TYPE_BIT_SIZE;

    /**
     * Number of actually used bits in <code>int</code> representation of sequence of types.
     */
    private static final int ENCODED_TYPE_LIST_BIT_SIZE =
        MAX_ENCODED_TYPE_LIST_LENGTH * Type.ENCODED_TYPE_BIT_SIZE;

    /**
     * Bit mask covering all bits used in <code>int</code> representation of sequence of types.
     */
    private static final int ENCODED_TYPE_LIST_BIT_MASK =
        ENCODED_TYPE_LIST_BIT_SIZE == IntUtils.INT_BITS_COUNT
        ? -1 : (1 << ENCODED_TYPE_LIST_BIT_SIZE) - 1;

    /**
     * The array contains multiplied encoded types for all possible types that can be
     * represented by {@link Type} class. Every element from the array contains multiple copies
     * of encoded single type packed adjacently to fill the whole word.
     */
    private static final int[] MULTIPLIED_ENCODED_TYPES;
    static {
        MULTIPLIED_ENCODED_TYPES = new int[1 << Type.ENCODED_TYPE_BIT_SIZE];
        for (int encType = 0; encType < MULTIPLIED_ENCODED_TYPES.length; encType++) {
            int value = 0;
            for (int j = 0; j < MAX_ENCODED_TYPE_LIST_LENGTH; j++) {
                value <<= Type.ENCODED_TYPE_BIT_SIZE;
                value |= encType;
            }
            MULTIPLIED_ENCODED_TYPES[encType] = value;
        }
    }

    /**
     * Verifies whether limits defined by {@link Config} class are not too loose.
     */
    static {
        if (Debugging.DEBUG_ENABLED) {
            Debugging.assume(
                Config.FUNC_REQ_PARAMS_COUNT_LIMIT < MAX_ENCODED_TYPE_LIST_LENGTH,
                "Limits.FUNC_REQ_PARAMS_COUNT_LIMIT must be less than "
                + MAX_ENCODED_TYPE_LIST_LENGTH);
        }
    }

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private EncodedTypeList() { }

    /**
     * Encodes given sequence of types and an extension into <code>long</code> value
     * representation.
     *
     * @param types
     *     array of {@link Type} instances to encode; its length may exceed actual number of
     *     types (specified by <code>typesCount</code> parameter), in this case only array's
     *     components with indices less than <code>typesCount</code> will be used.
     * @param typesCount actual number of types in <code>types</code> array (positive or zero)
     * @param extType "extension" of sequence of types
     * @return
     *     representation of a sequence of types and an extension in a form of <code>long</code>
     *     value
     */
    static long encodeTypeList(Type[] types, int typesCount, Type extType) {
        Debugging.assume(
            typesCount < MAX_ENCODED_TYPE_LIST_LENGTH,
            "Too many types to encode, maximum allowed number of types: "
            + (MAX_ENCODED_TYPE_LIST_LENGTH - 1));
        int encTypes = MULTIPLIED_ENCODED_TYPES[extType.getEncodedType()];
        for (int i = typesCount - 1; i >= 0; i--) {
            encTypes <<= Type.ENCODED_TYPE_BIT_SIZE;
            encTypes |= types[i].getEncodedType();
        }
        encTypes &= ENCODED_TYPE_LIST_BIT_MASK;
        int encTypesMask = IntUtils.getLowestBitsSet(typesCount * Type.ENCODED_TYPE_BIT_SIZE);
        long longTypesPart = encTypes;
        long longMaskPart = encTypesMask;
        return (longMaskPart << IntUtils.INT_BITS_COUNT)
               | (longTypesPart & (-1L >>> IntUtils.INT_BITS_COUNT));
    }

    /**
     * Checks whether given sequence of types of actual values matches given sequence of
     * declared types (with an extension). The method may be used to check if a function with
     * given declared types of parameters can accept actual parameters with given types.
     *
     * @param declEncTypeList
     *     encoded sequence of types representing declared types together with an extension
     * @param actEncTypeList
     *     encoded sequence of types representing types of actual values (it shouldn't have an
     *     extension)
     *
     * @return
     *     <code>true</code> if given sequence of types of actual values matches sequence of
     *     declared types
     */
    static boolean isTypeListMatching(long declEncTypeList, long actEncTypeList) {
        int masksSum = getMaskPart(declEncTypeList) | getMaskPart(actEncTypeList);
        return (getTypePart(declEncTypeList) & masksSum)
               == (getTypePart(actEncTypeList) & masksSum);
    }

    /**
     * Checks whether given sequence of types of actual values matches one of given sequences
     * of declared types (with extensions) and gives an information about matching sequence.
     * The method may be used to search a group of functions with given declared types of
     * parameters for single function that can accept actual parameters with given types.
     *
     * @param declEncTypeLists
     *     array of encoded sequences of types of declared parameters (with extensions)
     * @param actEncTypeList
     *     sequence of types of actual values to match against sequences of declared parameters
     * @return
     *     nonnegative index of encoded sequence of types of declared parameters from
     *     <code>declEncTypeLists</code> array that matches given encoded sequence of types of
     *     actual values (<code>actEncTypeList</code>) or <code>-1</code> if none of sequences
     *     matches
     */
    static int findMatchingTypeList(long[] declEncTypeLists, long actEncTypeList) {
        int actMask = getMaskPart(actEncTypeList);
        int actType = getTypePart(actEncTypeList);
        for (int i = 0; i < declEncTypeLists.length; i++) {
            long declEncTypeList = declEncTypeLists[i];
            int masksSum = getMaskPart(declEncTypeList) | actMask;
            if ((getTypePart(declEncTypeList) & masksSum) == (actType & masksSum)) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Returns part of encoded sequence of types that contains only packed encoded types from
     * sequence.
     *
     * @param encTypes
     *     encoded sequence of types (with optional extension), as generated by
     *     {@link #encodeTypeList(Type[], int, Type)} method
     *     methods.
     * @return
     *    part of encoded sequence of types that contains packed encoded types from sequence and
     *    an extension, without any additional bits set
     */
    private static int getTypePart(long encTypes) {
        return (int) encTypes;
    }

    /**
     * Returns part of encoded sequence of types that contains only bit mask for packed encoded
     * types from sequence.
     *
     * @param encTypes
     *     encoded sequence of types (with optional extension), as generated by
     *     {@link #encodeTypeList(Type[], int, Type)} method
     *     methods.
     * @return
     *    part of encoded sequence of types that contains bit mask covering bits from part
     *    containing encoded types (as returned by {@link #getTypePart(long)}) that describe
     *    only types from proper sequence, without extension
     */
    private static int getMaskPart(long encTypes) {
        return (int) (encTypes >>> IntUtils.INT_BITS_COUNT);
    }

}
