/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.repo;

import java.util.Arrays;
import pl.chyla.wyjec.Config;
import pl.chyla.wyjec.core.Function;
import pl.chyla.wyjec.core.Type;
import pl.chyla.wyjec.utils.ArrayUtils;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Repository of {@link pl.chyla.wyjec.core.Function} objects. A function may be fetched using
 * its name and an array of types of accepted parameters (overloading of functions' names is
 * supported).
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FunctionsRepository {

    /**
     * Second-level arrays contain names of functions encoded to <code>long</code> numbers and
     * sorted numerically.
     * <p>
     * First-level array is used to split numbers representing functions' names by names'
     * lengths. Sub-array at index n contains encoded names with length n.
     * <p>
     * Every component of <code>funcsEncNamesByLength</code> array has its counterpart in
     * <code>funcsGroupsByNameLength</code> array. Number stored as
     * <code>funcsEncNamesByLength[i][j]</code> represents encoded name of functions' group
     * stored as <code>funcsGroupsByNameLength[i][j]</code>
     */
    private long[][] funcsEncNamesByLength;

    /**
     * Second-level arrays contain groups of functions sorted by names (every group contains
     * functions with the same name, name of a group is defined as common name of all contained
     * functions).
     * <p>
     * First-level array is used to split functions' groups by lengths of their names.
     * Sub-array at index n contains groups which names have length n.
     */
    private FunctionsGroup[][] funcsGroupsByNameLength;

    /**
     * Creates repository of functions using given array of functions to initialize its
     * contents.
     * <p>
     * Functions' repository puts some limitations on valid names of functions. A function's
     * name is valid if and only if:
     * <ul>
     * <li>it's length is between 1 and 9 (<code>MAX_FUNCTION_NAME_LENGTH</code>) inclusive</li>
     * and
     * <li>it consists only of ASCII characters</li>
     * </ul>
     *
     * @param funcs array of functions to put in the repository
     * @throws IllegalArgumentException
     *     in case <code>funcs</code> array contains a function with invalid name or with too
     *     many declared required parameters (&gt; {@link Config#FUNC_REQ_PARAMS_COUNT_LIMIT})
     *     or it contains two functions that can accept the same list of actual parameters.
     */
    public FunctionsRepository(Function[] funcs) {
        checkFunctions(funcs);
        FunctionsGroup[] funcsGroups = makeSortedFunctionsGroups(funcs);
        // allocate arrays funcsEncNamesByLength and funcsGroupsByNameLength
        funcsEncNamesByLength = new long[Config.FUNC_NAME_LENGTH_LIMIT + 1][];
        funcsGroupsByNameLength = new FunctionsGroup[Config.FUNC_NAME_LENGTH_LIMIT + 1][];
        int[] countsByLength = new int[Config.FUNC_NAME_LENGTH_LIMIT + 1];
        Arrays.fill(countsByLength, 0);
        for (int i = 0; i < funcsGroups.length; i++) {
            int len = funcsGroups[i].getName().length();
            countsByLength[len]++;
        }
        for (int i = 0; i <= Config.FUNC_NAME_LENGTH_LIMIT; i++) {
            int count = countsByLength[i];
            funcsEncNamesByLength[i] = new long[count];
            funcsGroupsByNameLength[i] = new FunctionsGroup[count];
        }
        // fill arrays funcsEncNamesByLength and funcsGroupsByNameLength
        Arrays.fill(countsByLength, 0);
        for (int i = 0; i < funcsGroups.length; i++) {
            int len = funcsGroups[i].getName().length();
            int count = countsByLength[len];
            String groupName = funcsGroups[i].getName();
            long encGroupName = FunctionNameEncoder.encodeName(new RawString(groupName));
            funcsEncNamesByLength[len][count] = encGroupName;
            funcsGroupsByNameLength[len][count] = funcsGroups[i];
            countsByLength[len]++;
        }
    }

    /**
     * Checks whether given array of functions is valid for use as contents of repository.
     *
     * @param funcs array of functions to check
     * @throws IllegalArgumentException
     *     in case <code>funcs</code> array contains a function with invalid name or with too
     *     many declared required parameters (&gt; {@link Config#FUNC_REQ_PARAMS_COUNT_LIMIT})
     */
    private void checkFunctions(Function[] funcs) {
        for (int i = 0; i < funcs.length; i++) {
            String name = funcs[i].getName();
            if (!FunctionNameEncoder.isValidName(new RawString(name))) {
                throw new IllegalArgumentException("Invalid function's name: " + name);
            }
            int reqParamsCount = funcs[i].getSignature().getRequiredParametersTypes().length;
            if (reqParamsCount > Config.FUNC_REQ_PARAMS_COUNT_LIMIT) {
                throw new IllegalArgumentException(
                    "Too many required parameters (" + reqParamsCount + ")"
                    + " used in signature of function " + name
                    + "(maximum allowed number of parameters is: "
                    + Config.FUNC_REQ_PARAMS_COUNT_LIMIT + ")");
            }
        }
    }

    /**
     * For given array of functions creates groups of functions with equal names as
     * {@link FunctionsGroup} objects and returns them in an array sorted lexicographically.
     *
     * @param funcs array of functions to process
     * @return
     *     sorted array of functions' groups containing all functions contained in
     *     <code>funcs</code> array
     */
    private FunctionsGroup[] makeSortedFunctionsGroups(Function[] funcs) {
        int funcsCount = funcs.length;
        // make copy of functions' table, sort it by functions' names
        Function[] tmpFuncs = new Function[funcsCount];
        ArrayUtils.copyArray(funcs, tmpFuncs);
        Arrays.sort(tmpFuncs, FunctionComparator.getInstance());
        // count number of functions' groups
        int groupsCount = 0;
        String prevName = "";
        for (int i = 0; i < funcsCount; i++) {
            String funcName = tmpFuncs[i].getName();
            if (!funcName.equals(prevName)) {
                groupsCount++;
                prevName = funcName;
            }
        }
        // create functions' groups
        FunctionsGroup[] funcsGroups = new FunctionsGroup[groupsCount];
        int fiGroupStart = 0;
        int fi = 0;
        int gi = 0;
        while (gi < groupsCount) {
            String groupName = tmpFuncs[fiGroupStart].getName();
            if (fi >= funcsCount || !tmpFuncs[fi].getName().equals(groupName)) {
                int groupLen = fi - fiGroupStart;
                Function[] group = new Function[groupLen];
                System.arraycopy(tmpFuncs, fiGroupStart, group, 0, groupLen);
                funcsGroups[gi] = FunctionsGroup.createGroup(group);
                gi++;
                fiGroupStart = fi;
            }
            fi++;
        }
        return funcsGroups;
    }

    /**
     * Searches the repository for function with given name that can accepts parameters with
     * given types. Constraints checked during creation of the repository guarantee that for
     * every name and parameters' types there will be at most one matching {@link Function}
     * object found.
     *
     * @param name name of a function to look for
     * @param paramsTypes
     *     array of parameters' types to match to types of declared parameters of a function;
     *     the array may contain additional unused values at the that will be ignored (only
     *     <code>paramsCount</code> first elements will be taken into account)
     * @param paramsCount
     *     number of actual parameters, must be less or equal to length of
     *     <code>paramsTypes</code> array
     * @return function object or <code>null</code> if no function can be found
     */
    public Function lookup(ReadableRawString name, Type[] paramsTypes, int paramsCount) {
        if (!FunctionNameEncoder.isValidName(name)) {
            return null;
        }
        long encName = FunctionNameEncoder.encodeName(name);
        int len = name.getLength();
        int groupIdx = Arrays.binarySearch(funcsEncNamesByLength[len], encName);
        Function func;
        if (groupIdx >= 0) {
            FunctionsGroup group = funcsGroupsByNameLength[len][groupIdx];
            func = group.lookupFunction(paramsTypes, paramsCount);
        } else {
            func = null;
        }
        return func;
    }

}
