/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdconsts;

import pl.chyla.wyjec.Debugging;
import pl.chyla.wyjec.core.IConstantsParser;
import pl.chyla.wyjec.core.Value;
import pl.chyla.wyjec.utils.CharUtils;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Parser for float constants with decimal representations matching regular expression
 * "[+-]?([0-9]+\.[0-9]+(e[+-]?[0-9]+)?)|([0-9]+e[+-]?[0-9]+)", e.g. <code>0.012</code>,
 * <code>12.34</code>, <code>007.80e-123</code>, <code>785e3</code>, <code>-12.765</code>,
 * <code>+13</code>.
 *
 * @see IConstantsParser
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FloatConstantsParser implements IConstantsParser {

    /**
     * Instance of <code>FloatConstantsParser</code>, accessible from outside via
     * {@link #getInstance()}.
     */
    private static final FloatConstantsParser instance = new FloatConstantsParser();

    /**
     * The maximum <code>long</code> number that consists only from '9' digits. It has 18
     * decimal digits.
     */
    private static final long MAX_SIGNIFICANT_NUMBER = 999999999999999999L;

    /**
     * Private (empty) constructor to disallow instantiating of the class from outside. Class'
     * instance should be accessed via {@link #getInstance()} getter.
     */
    private FloatConstantsParser() { }

    /**
     * Returns instance of <code>FloatConstantsParser</code> which can be safely used independly
     * from other threads which also get <code>FloatConstantsParser</code> using this method.
     * It's undefined whether this method returns newly created instance or not.
     *
     * @return instance of <code>FloatConstantsParser</code>
     */
    public static FloatConstantsParser getInstance() {
        return instance;
    }

    /**
     * Parses given string as a representation of float constant value. If the string contains a
     * representation of float matching regular expression
     * "[+-]?([0-9]+\.[0-9]+(e[+-]?[0-9]+)?)|([0-9]+e[+-]?[0-9]+)" its numeric value will be
     * stored in given {@link Value} and can be later accessed using {@link Value#getFloat()}
     * method. The returned value will be either finite number or infinity (positive or
     * negative).
     *
     * @param result {@link Value} object used to store recognized float value
     * @param constStr string representation of constant to parse
     * @return
     *     <code>true</code> if given string contains decimal representation of float number (in
     *     this case its value can be accessed using given {@link Value} object),
     *     <code>false</code> otherwise.
     * @see IConstantsParser#parseConstant(Value, ReadableRawString)
     */
    public boolean parseConstant(Value result, ReadableRawString constStr) {
        char[] chars = constStr.getStorage();
        int startPos = constStr.getStartPos();
        int endPos = constStr.getEndPos();
        if (!isValidFloat(chars, startPos, endPos)) {
            return false;
        }
        int curPos = startPos;
        double sign;
        if (chars[curPos] == '+') {
            sign = 1.0;
            curPos++;
        } else if (chars[curPos] == '-') {
            sign = -1.0;
            curPos++;
        } else {
            sign = 1.0;
        }
        double value = parseSimplePart(chars, curPos, endPos);
        while (curPos < endPos && chars[curPos] != 'e') {
            curPos++;
        }
        if (value != 0.0 && curPos < endPos) {
            double exponent = parseExponentPart(chars, curPos, endPos);
            value *= Math.pow(10.0, exponent);
        }
        result.setFloat(sign * value);
        return true;
    }

    /**
     * Parses "simple" form of float constant, i.e. without sign and exponent part. Parsed form
     * must match regular expression "[0-9]+(\.[0-9]+)?". This method doesn't validate input at
     * all so its correctness must be verified earlier. Characters following simple form of
     * float will not be parsed (parsing will stop before <code>endPos</code> or after parsing
     * the last character belonging to simple form).
     *
     * @param chars array containing sequence of characters to parse
     * @param startPos position of first character to parse in the array
     * @param endPos position after the last character to parse in the array
     * @return parsed value (nonnegative finite number)
     */
    private double parseSimplePart(char[] chars, int startPos, int endPos) {
        long value = chars[startPos] - '0';
        int curPos = startPos + 1;
        while (CharUtils.isDigit(chars[curPos]) && value <= MAX_SIGNIFICANT_NUMBER / 10) {
            value = value * 10L + (chars[curPos] - '0');
            curPos++;
        }
        double doubleValue;
        double valueMultiplier = 1.0;
        if (chars[curPos] == '.') {
            curPos++;
            while (curPos < endPos && CharUtils.isDigit(chars[curPos])
                   && value <= MAX_SIGNIFICANT_NUMBER / 10) {
                value = value * 10L + (chars[curPos] - '0');
                valueMultiplier *= 10.0;
                curPos++;
            }
            doubleValue = value / valueMultiplier;
        } else {
            while (CharUtils.isDigit(chars[curPos])) {
                valueMultiplier *= 10.0;
                curPos++;
            }
            doubleValue = value * valueMultiplier;
        }
        return doubleValue;
    }

    /**
     * Parses exponent part of float constant, i.e. the part starting with "e" character. This
     * method doesn't validate input at all and assumes that exponent part actually exists, so
     * verification must be performed earlier (this implies that character at
     * <code>startPos</code> must be "e").
     *
     * @param chars array containing sequence of characters to parse
     * @param startPos position of first character to parse in the array
     * @param endPos position after the last character to parse in the array
     * @return parsed exponent value of <code>0.0</code> if exponent part is not present
     */
    private double parseExponentPart(char[] chars, int startPos, int endPos) {
        Debugging.assume(
            startPos < endPos && chars[startPos] == 'e',
            "Exponent part should start with 'e' character");
        int curPos = startPos + 1;
        double sign;
        if (chars[curPos] == '+') {
            sign = 1.0;
            curPos++;
        } else if (chars[curPos] == '-') {
            sign = -1.0;
            curPos++;
        } else {
            sign = 1.0;
        }
        double exponent = 0.0;
        do {
            exponent = exponent * 10.0 + (chars[curPos] - '0');
            curPos++;
        } while (curPos < endPos);
        exponent *= sign;
        return exponent;
    }

    /**
     * Checks whether given sequence of characters (specified as part of given array) contains
     * valid float constant. Valid float constant matches the following regular expression:
     * "[+-]?([0-9]+\.[0-9]+(e[+-]?[0-9]+)?)|([0-9]+e[+-]?[0-9]+)".
     *
     * @param chars array containing sequence of characters to check
     * @param startPos position of first character to check in the array
     * @param endPos position after the last character to check in the array
     * @return
     *     <code>true</code> if specified sequence of characters contains valid float constant,
     *     <code>false</code> otherwise.
     */
    private boolean isValidFloat(char[] chars, int startPos, int endPos) {
        int curPos = startPos;
        if (curPos < endPos && (chars[curPos] == '+' || chars[curPos] == '-')) {
            curPos++;
        }
        int intPartSize = getIntegerPartSize(chars, curPos, endPos);
        curPos += intPartSize;
        int fractPartSize = getFractionalPartSize(chars, curPos, endPos);
        curPos += fractPartSize;
        int expPartSize = getExponentPartSize(chars, curPos, endPos);
        curPos += expPartSize;
        return curPos == endPos && intPartSize > 0 && fractPartSize + expPartSize > 0;
    }

    /**
     * Evaluates the length of integer part of float constant, i.e. the number of consecutive
     * digits starting from given position. Counting stops at any non-digit character or after
     * reaching <code>endPos</code>.
     *
     * @param chars array containing sequence of characters to scan
     * @param startPos position of first character to scan
     * @param endPos position after the last character to scan
     * @return the length of integer part of float constant (nonnegative number)
     */
    private int getIntegerPartSize(char[] chars, int startPos, int endPos) {
        int curPos = startPos;
        while (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
            curPos++;
        }
        return curPos - startPos;
    }

    /**
     * Evaluates the length of optional fractional part of float constant, i.e. the number of
     * characters matching regular expression "\.[0-9]+". If sequence of characters at given
     * position doesn't match required pattern, <code>0</code> is returned.
     *
     * @param chars array containing sequence of characters to scan
     * @param startPos
     *     position of first character to scan (position immediately after integer part)
     * @param endPos position after the last character to scan
     * @return the length of fractional part or <code>0</code> if fractional part doesn't exist
     */
    private int getFractionalPartSize(char[] chars, int startPos, int endPos) {
        boolean isValid = false;
        int curPos = startPos;
        if (curPos < endPos && chars[curPos] == '.') {
            curPos++;
            if (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
                isValid = true;
                curPos++;
                while (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
                    curPos++;
                }
            }
        }
        if (isValid) {
            return curPos - startPos;
        } else {
            return 0;
        }
    }

    /**
     * Evaluates the length of optional exponent part of float constant, i.e. the number of
     * characters matching regular expression "e[+-]?[0-9]+". If sequence of characters at given
     * position doesn't match required pattern, <code>0</code> is returned.
     *
     * @param chars array containing sequence of characters to scan
     * @param startPos
     *     position of first character to scan (position immediately after fractional part or
     *     after integer part if fractional part doesn't exist)
     * @param endPos position after the last character to scan
     * @return the length of fractional part or <code>0</code> if fractional part doesn't exist
     */
    private int getExponentPartSize(char[] chars, int startPos, int endPos) {
        boolean isValid = false;
        int curPos = startPos;
        if (curPos < endPos && chars[curPos] == 'e') {
            curPos++;
            if (curPos < endPos && (chars[curPos] == '-' || chars[curPos] == '+')) {
                curPos++;
            }
            if (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
                curPos++;
                isValid = true;
                while (curPos < endPos && CharUtils.isDigit(chars[curPos])) {
                    curPos++;
                }
            }
        }
        if (isValid) {
            return curPos - startPos;
        } else {
            return 0;
        }
    }

}
