/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.datefuncs.selection;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.DateFunction;
import pl.chyla.wyjec.core.optparameters.DateOptParameters;
import pl.chyla.wyjec.core.parameters.DateParameter;
import pl.chyla.wyjec.core.parameters.IntegerParameter;


/**
 * Returns one of given dates depending on value of selector. Selector specifies 0-based
 * index of date from given sequence of optional dates. In case index is negative or too
 * big, default date is returned.
 *
 * @wyjec.func select
 * @wyjec.example
 * "{select|{&#64;2005-01-01}|{1}|{&#64;1978-09-11}|{&#64;1977-03-09}|{&#64;1975-04-09}}"
 * "1977-03-09 00:00:00" date
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnSelect extends DateFunction {

    /**
     * Default date used as result of the function in case selector is negative or greater
     * than length of sequence of dates
     * @wyjec.param defaultDate
     */
    private static final DateParameter pDefaultDate = new DateParameter();

    /**
     * Index of value from sequence of dates to return as result
     * @wyjec.param selector
     */
    private static final IntegerParameter pSelector = new IntegerParameter();

    /**
     * Optional sequence of dates to select single value from (may be empty)
     * @wyjec.param optDates
     */
    private static final DateOptParameters pOptDates = new DateOptParameters();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pDefaultDate, pSelector}, pOptDates, true);

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "select";
    }

    protected long execute() {
        int selector = (int) getParameter(pSelector);
        if (selector >= 0 && selector < getOptParametersCount(pOptDates)) {
            return getParameter(pOptDates, selector);
        } else {
            return getParameter(pDefaultDate);
        }
    }

}
