/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.stringfuncs.ops;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.IntegerParameter;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Returns a copy of given string value, prefixing it with padding if its length is smaller than
 * the specified one. Characters from given padding will be repeated if necessary, up to the
 * required length. If given padding is empty, the function will return unmodified string.
 *
 * @wyjec.func lpad
 * @wyjec.example "{lpad|test|{10}|._}" "._._._test" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnLPad extends StringFunction {

    /**
     * Base string used to generate the new one
     * @wyjec.param value
     */
    private static final StringParameter pValue = new StringParameter();

    /**
     * Required minimum length of the resulting string.
     * @wyjec.param length
     */
    private static final IntegerParameter pLength = new IntegerParameter();

    /**
     * String to use for padding.
     * @wyjec.param padding
     */
    private static final StringParameter pPadding = new StringParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pValue, pLength, pPadding}, null);

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "lpad";
    }

    protected void execute(RawString result) {
        ReadableRawString value = getParameter(pValue);
        ReadableRawString padding = getParameter(pPadding);
        if (padding.getLength() != 0) {
            int spaceLeft = (int) Math.max(getParameter(pLength) - value.getLength(), 0);
            while (spaceLeft >= padding.getLength()) {
                result.append(padding);
                spaceLeft -= padding.getLength();
            }
            if (spaceLeft > 0) {
                result.append(padding, 0, spaceLeft);
            }
        }
        result.append(value);
    }

}
