/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.stdfuncs.stringfuncs.ops;

import pl.chyla.wyjec.core.FunctionSignature;
import pl.chyla.wyjec.core.Parameter;
import pl.chyla.wyjec.core.functions.StringFunction;
import pl.chyla.wyjec.core.parameters.StringParameter;
import pl.chyla.wyjec.utils.RawString;
import pl.chyla.wyjec.utils.ReadableRawString;


/**
 * Returns modified copy of given string, such as all non-overlapping occurrences of another
 * given string are replaced with specified replacement. If substring to be replaced is empty,
 * the function will return unmodified string.
 *
 * @wyjec.func replace
 * @wyjec.example "{replace|free culture|e|**}" "fr**** cultur**" string
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class FnReplace extends StringFunction {

    /**
     * Base string to perform replacement on
     * @wyjec.param value
     */
    private static final StringParameter pValue = new StringParameter();

    /**
     * Old substring to search for in the base string
     * @wyjec.param oldSub
     */
    private static final StringParameter pOldSub = new StringParameter();

    /**
     * New substring to use as a replacement for occurrences of old substring
     * @wyjec.param newSub
     */
    private static final StringParameter pNewSub = new StringParameter();

    private static final FunctionSignature signature = new FunctionSignature(
        new Parameter[] {pValue, pOldSub, pNewSub}, null);

    public FunctionSignature getSignature() {
        return signature;
    }

    public String getName() {
        return "replace";
    }

    protected void execute(RawString result) {
        ReadableRawString value = getParameter(pValue);
        ReadableRawString oldSub = getParameter(pOldSub);
        ReadableRawString newSub = getParameter(pNewSub);
        int oldSubLen = oldSub.getLength();
        if (oldSubLen == 0) {
            result.append(value);
            return;
        }
        int curPos = 0;
        while (curPos < value.getLength()) {
            int subPos = value.find(oldSub, curPos);
            if (subPos != -1) {
                result.append(value, curPos, subPos - curPos);
                result.append(newSub);
                curPos = subPos + oldSubLen;
            } else {
                result.append(value, curPos, value.getLength() - curPos);
                break;
            }
        }
    }

}
