/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;


/**
 * Utility class allowing for fast writing of string representation of <code>double</code>
 * number into existing array of characters.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class DoubleFormatter {

    /**
     * Maximum number of characters (including sign character and other optional characters) for
     * string representation of number of type <code>double</code>
     */
    public static final int MAX_CHARACTERS_COUNT = "-9.99999999999999e-999".length();

    /**
     * Maximum number of decimal digits that can be stored in <code>double</code> number without
     * losing precision.
     */
    private static final int SIGNIFICANT_DIGITS_COUNT = 15;

    /**
     * Maximum number that can be represented with significant digits, i.e.
     * <code>10 ** {@link #SIGNIFICANT_DIGITS_COUNT} - 1</code>
     */
    private static final long MAX_WITH_SIGNIFICANT_DIGITS = 999999999999999L;

    /**
     * String representing special zero double value (0.0)
     */
    private static final String ZERO_STRING = "0.0";

    /**
     * String representing special double value: Not a Number.
     */
    private static final String NAN_STRING = "NaN";

    /**
     * String representing special double value: infinity.
     */
    private static final String INFINITY_STRING = "inf";

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private DoubleFormatter() { }

    /**
     * Stores string representation of given number in consecutive components of specified array
     * of characters, using scientific notation or simple notation with decimal point.
     * Characters will be written in the array starting from specified index.
     *
     * @param value value to generate string representation for
     * @param dest array to write string representation to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if string representation doesn't fit
     *     in specified part of the array (no character is written in this case)
     */
    public static int formatToArray(double value, char[] dest, int startPos) {
        int storedCharsCount;
        if (Double.isNaN(value)) {
            storedCharsCount = formatStringToArray(NAN_STRING, dest, startPos);
        } else if (Math.abs(value) < Double.MIN_VALUE) {
            storedCharsCount = formatStringToArray(ZERO_STRING, dest, startPos);
        } else if (value < 0) {
            int posCharsCount = formatPositive(-value, dest, startPos + 1);
            if (posCharsCount > 0) {
                dest[startPos] = '-';
                storedCharsCount = posCharsCount + 1;
            } else {
                storedCharsCount = 0;
            }
        } else {
            storedCharsCount = formatPositive(value, dest, startPos);
        }
        return storedCharsCount;
    }

    /**
     * Stores string representation of given positive number in specified part of array.
     *
     * @param value
     *     value to generate string representation for positive finite value or positive
     *     infinity
     * @param dest array to write string representation to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if string representation doesn't fit
     *     in specified part of the array (no character is written in this case)
     */
    private static int formatPositive(double value, char[] dest, int startPos) {
        int storedCharsCount;
        if (Double.isInfinite(value)) {
            storedCharsCount = formatStringToArray(INFINITY_STRING, dest, startPos);
        } else {
            if (Math.floor(value) <= MAX_WITH_SIGNIFICANT_DIGITS / 10 && value >= 0.1) {
                storedCharsCount = formatSimplePositiveValue(value, dest, startPos);
            } else {
                storedCharsCount = formatScientificPositiveValue(value, dest, startPos);
            }
        }
        return storedCharsCount;
    }

    /**
     * Stores string representation of given positive number in specified part of array,
     * using scientific notation, i.e. [number]e[exponent] (exponent may be negative).
     *
     * @param value value to generate string representation for
     * @param dest array to write string representation to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if string representation doesn't fit
     *     in specified part of the array (no character is written in this case)
     */
    private static int formatScientificPositiveValue(double value, char[] dest, int startPos) {
        int freeChars = dest.length - startPos;
        int shift = (int) Math.ceil(log10(value)) - SIGNIFICANT_DIGITS_COUNT;
        long shiftedValue = (long) (value / Math.pow(10.0, shift));
        while (shiftedValue > MAX_WITH_SIGNIFICANT_DIGITS) {
            shiftedValue /= 10L;
            shift++;
        }
        while (shiftedValue % 10L == 0L) {
            shiftedValue /= 10L;
            shift++;
        }
        int charsCount;
        int mantissaDigitsCount = LongFormatter.getCharactersCount(shiftedValue);
        if (mantissaDigitsCount == 1) {
            charsCount = mantissaDigitsCount + 1 + IntFormatter.getCharactersCount(shift);
        } else {
            shift += mantissaDigitsCount - 1;
            charsCount = mantissaDigitsCount + 1 + 1 + IntFormatter.getCharactersCount(shift);
        }
        if (charsCount <= freeChars) {
            int curPos = startPos;
            if (mantissaDigitsCount > 1) {
                LongFormatter.formatToArray(shiftedValue, dest, curPos + 1);
                dest[curPos] = dest[curPos + 1];
                dest[curPos + 1] = '.';
                curPos += mantissaDigitsCount + 1;
            } else {
                dest[curPos] = (char) (shiftedValue + '0');
                curPos++;
            }
            dest[curPos] = 'e';
            curPos++;
            curPos += IntFormatter.formatToArray(shift, dest, curPos);
            return charsCount;
        } else {
            return 0;
        }
    }

    /**
     * Stores string representation of given positive number in specified part of array,
     * using simple notation with decimal dot, i.e. [digits].[digits]. This method supports
     * only numbers with integral part consisting of less than {@link #SIGNIFICANT_DIGITS_COUNT}
     * digits. It's not recommended to use this method for values less than <code>0.1</code>,
     * because it will cause losing precision in output string representation.
     *
     * @param value value to generate string representation for
     * @param dest array to write string representation to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if string representation doesn't fit
     *     in specified part of the array (no character is written in this case)
     * @see #formatScientificPositiveValue(double, char[], int)
     */
    private static int formatSimplePositiveValue(double value, char[] dest, int startPos) {
        int freeChars = dest.length - startPos;
        long intDigits = (long) value;
        int intDigitsCount = LongFormatter.getCharactersCount(intDigits);
        int fractDigitsCount = SIGNIFICANT_DIGITS_COUNT - intDigitsCount;
        long fractDigits = (long) ((value - Math.floor(value)) * pow10(fractDigitsCount));
        while (fractDigitsCount > 1 && fractDigits % 10L == 0L) {
            fractDigits /= 10L;
            fractDigitsCount--;
        }
        int charsCount = intDigitsCount + 1 + fractDigitsCount;
        if (charsCount <= freeChars) {
            int curPos = startPos;
            LongFormatter.formatToArray(intDigits, dest, curPos);
            curPos += intDigitsCount;
            dest[curPos] = '.';
            curPos++;
            for (int i = fractDigitsCount - 1; i >= 0; i--) {
                dest[curPos + i] = (char) ('0' + fractDigits % 10L);
                fractDigits /= 10L;
            }
            return charsCount;
        } else {
            return 0;
        }
    }

    /**
     * Stores given string in specified part of array.
     *
     * @param strValue value to store in the array
     * @param dest array to write given string to
     * @param startPos
     *     index of first component of <code>dest</code> array to use for writing characters
     * @return
     *     number of characters written to the array or 0 if given string doesn't fit in
     *     specified part of the array (no character is written in this case)
     */
    private static int formatStringToArray(String strValue, char[] dest, int startPos) {
        int freeChars = dest.length - startPos;
        if (strValue.length() <= freeChars) {
            strValue.getChars(0, strValue.length(), dest, startPos);
            return strValue.length();
        } else {
            return 0;
        }
    }

    /**
     * Returns logarithm with base 10 of given <code>double</code> value.
     *
     * @param value value to calculate the logarithm for
     * @return logarithm with base 10 of <code>value</code>
     */
    private static double log10(double value) {
        return Math.log(value) / Math.log(10.0);
    }

    /**
     * Returns <code>10</code> raised to the power of given parameter as <code>long</code>
     * number.
     *
     * @param exp the exponent to use
     * @return power of <code>10</code> with given exponent
     */
    private static long pow10(int exp) {
        long result = 1;
        for (int i = 0; i < exp; i++) {
            result *= 10;
        }
        return result;
    }

}
