/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;


/**
 * Helper class grouping simple (and effective) operations on numbers of type <code>int</code>.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class IntUtils {

    /**
     * Number of bits in values of type int.
     */
    public static final int INT_BITS_COUNT = 32;

    /**
     * Constant equal to <code>log2(INT_BITS_COUNT)</code>
     */
    private static final int INT_BITS_COUNT_LOG2 = 5;

    /**
     * Private empty constructor to disallow instantiating of the class.
     */
    private IntUtils() { }

    /**
     * Generates integer number with only given number of most significant bits set.
     *
     * @param bitsCount number of most significant bits to set (0 &lt;= bitsCount &lt;= 32)
     * @return number with only <code>bitsCount</code> of highest bits set
     */
    public static int getWithHighestBitsSet(int bitsCount) {
        return ~getLowestBitsSet(INT_BITS_COUNT - bitsCount);
    }

    /**
     * Generates integer number with only given number of least significant bits set.
     *
     * @param bitsCount number of least significant bits to set (0 &lt;= bitsCount &lt;= 32)
     * @return number with only <code>bitsCount</code> of lowest bits set
     */
    public static int getLowestBitsSet(int bitsCount) {
        int bitsCountHighestBit = (bitsCount >>> INT_BITS_COUNT_LOG2) & 0x01;
        return (1 << bitsCount) - 1 - bitsCountHighestBit;
    }

    /**
     * Checks whether given <code>value</code> fits in range &lt;<code>min</code>,
     * <code>max</code>&gt; (inclusive). Value of <code>min</code> must not be greater than
     * <code>max</code>.
     * <p>
     * <strong>Note about used optimization technique:</strong>
     * <p>
     * the following expression:
     * <p>
     * <blockquote>
     * <code>Integer.MIN_VALUE + value - min &lt;= Integer.MIN_VALUE + (max - min)<code>
     * </blockquote>
     * is an equivalent for:
     * <blockquote>
     * <code>value &gt;= min &amp;&amp; digitValue &lt;= max</code>
     * </blockquote>
     *
     * @param value value to check
     * @param min lower limit of the range
     * @param max upper limit of the range
     * @return
     *     <code>true</code> if <code>min</code> &lt;= <code>value</code> &lt;=
     *     <code>max</code>, <code>false</code> otherwise
     */
    public static boolean isBetweenInclusive(int value, int min, int max) {
        return Integer.MIN_VALUE + value - min <= Integer.MIN_VALUE + (max - min);
    }

    /**
     * Compares two numbers for order. It's direct equivalent of
     * {@link java.util.Comparator#compare(java.lang.Object, java.lang.Object)} method but
     * operates on primitive values of type <code>int</code>.
     *
     * @param a first number to be compared
     * @param b second number to be compared
     * @return
     *     negative number, zero or positive number if first number is less, equal or greater
     *     than the second one.
     */
    public static int compare(int a, int b) {
        int result;
        if (a < b) {
            result = -1;
        } else if (a > b) {
            result = 1;
        } else {
            result = 0;
        }
        return result;
    }

}
