/*
 * Wyjec - expression language, version 1.0
 * (c) Copyright by Zbigniew Chyla 2005
 *
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License version 2.1 as published by the Free Software
 * Foundation.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301  USA
 */
package pl.chyla.wyjec.utils;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.TimeZone;


/**
 * Subclass of {@link java.util.GregorianCalendar} making it a bit simpler to use, i.e. its
 * constructor has more predictable behaviour (always uses GMT time zone and English locale)
 * and the class provides extra methods for setting date and time.
 *
 * @author Zbigniew Chyla (zbigniew@chyla.pl)
 */
public final class SimpleCalendar extends GregorianCalendar {

    /**
     * Constructs <code>SimpleCalendar</code> instance using GMT time zone and English locale.
     */
    public SimpleCalendar() {
        super(TimeZone.getTimeZone("GMT"), Locale.ENGLISH);
    }

    /**
     * Constructs <code>SimpleCalendar</code> instance using specified time zone and locale.
     *
     * @param timeZone time zone to use
     * @param locale to use
     */
    public SimpleCalendar(TimeZone timeZone, Locale locale) {
        super(timeZone, locale);
    }

    /**
     * Sets fields in this calendar to midnight of given date.
     *
     * @param year year number
     * @param month 0-based month of the year
     * @param day 1-based day of the month
     */
    public void setDate(int year, int month, int day) {
        setDate(year, month, day, 0, 0, 0);
    }

    /**
     * Sets fields in this calendar to given date and time.
     *
     * @param year year number
     * @param month 0-based month of the year
     * @param day 1-based day of the month
     * @param hour 0-based hour of the day
     * @param minute 0-based minute of the hour
     * @param second 0-based second of the minute
     */
    public void setDate(int year, int month, int day, int hour, int minute, int second) {
        set(year, month, day, hour, minute, second);
        set(Calendar.MILLISECOND, 0);
    }

}
